@@ -6,7 +6,7 @@
 
 from __future__ import absolute_import, print_function
 
-import base64
+import notebook
 import binascii
 import datetime
 import errno
@@ -23,8 +23,15 @@ import signal
 import socket
 import sys
 import threading
+import warnings
 import webbrowser
 
+try: #PY3
+    from base64 import encodebytes
+except ImportError: #PY2
+    from base64 import encodestring as encodebytes
+
+
 from jinja2 import Environment, FileSystemLoader
 
 # Install the pyzmq ioloop. This has to be done before anything else from
@@ -56,35 +63,44 @@ from notebook import (
     __version__,
 )
 
+# py23 compatibility
+try:
+    raw_input = raw_input
+except NameError:
+    raw_input = input
+
 from .base.handlers import Template404, RedirectWithParams
 from .log import log_request
 from .services.kernels.kernelmanager import MappingKernelManager
 from .services.config import ConfigManager
 from .services.contents.manager import ContentsManager
 from .services.contents.filemanager import FileContentsManager
+from .services.contents.largefilemanager import LargeFileManager
 from .services.sessions.sessionmanager import SessionManager
 
 from .auth.login import LoginHandler
 from .auth.logout import LogoutHandler
-from .base.handlers import FileFindHandler, IPythonHandler
+from .base.handlers import FileFindHandler
 
 from traitlets.config import Config
 from traitlets.config.application import catch_config_error, boolean_flag
 from jupyter_core.application import (
     JupyterApp, base_flags, base_aliases,
 )
+from jupyter_core.paths import jupyter_config_path
 from jupyter_client import KernelManager
 from jupyter_client.kernelspec import KernelSpecManager, NoSuchKernel, NATIVE_KERNEL_NAME
 from jupyter_client.session import Session
 from nbformat.sign import NotebookNotary
 from traitlets import (
     Dict, Unicode, Integer, List, Bool, Bytes, Instance,
-    TraitError, Type, Float,
+    TraitError, Type, Float, observe, default, validate
 )
 from ipython_genutils import py3compat
 from jupyter_core.paths import jupyter_runtime_dir, jupyter_path
 from notebook._sysinfo import get_sys_info
 
+from ._tz import utcnow
 from .utils import url_path_join, check_pid, url_escape
 
 #-----------------------------------------------------------------------------
@@ -94,6 +110,16 @@ from .utils import url_path_join, check_pid, url_escape
 _examples = """
 jupyter notebook                       # start the notebook
 jupyter notebook --certfile=mycert.pem # use SSL/TLS certificate
+jupyter notebook password              # enter a password to protect the server
+"""
+
+DEV_NOTE_NPM = """It looks like you're running the notebook from source.
+If you're working on the Javascript of the notebook, try running
+
+    npm run build:watch
+
+in another terminal window to have the system incrementally
+watch and build the notebook's JavaScript for you, as you make changes.
 """
 
 #-----------------------------------------------------------------------------
@@ -117,27 +143,33 @@ def load_handlers(name):
     mod = __import__(name, fromlist=['default_handlers'])
     return mod.default_handlers
 
-
 #-----------------------------------------------------------------------------
 # The Tornado web application
 #-----------------------------------------------------------------------------
 
 class NotebookWebApplication(web.Application):
 
-    def __init__(self, ipython_app, kernel_manager, contents_manager,
+    def __init__(self, jupyter_app, kernel_manager, contents_manager,
                  session_manager, kernel_spec_manager,
                  config_manager, log,
                  base_url, default_url, settings_overrides, jinja_env_options):
 
+        # If the user is running the notebook in a git directory, make the assumption
+        # that this is a dev install and suggest to the developer `npm run build:watch`.
+        base_dir = os.path.realpath(os.path.join(__file__, '..', '..'))
+        dev_mode = os.path.exists(os.path.join(base_dir, '.git'))
+        if dev_mode:
+            log.info(DEV_NOTE_NPM)
+
         settings = self.init_settings(
-            ipython_app, kernel_manager, contents_manager,
+            jupyter_app, kernel_manager, contents_manager,
             session_manager, kernel_spec_manager, config_manager, log, base_url,
             default_url, settings_overrides, jinja_env_options)
         handlers = self.init_handlers(settings)
 
         super(NotebookWebApplication, self).__init__(handlers, **settings)
 
-    def init_settings(self, ipython_app, kernel_manager, contents_manager,
+    def init_settings(self, jupyter_app, kernel_manager, contents_manager,
                       session_manager, kernel_spec_manager,
                       config_manager,
                       log, base_url, default_url, settings_overrides,
@@ -145,7 +177,7 @@ class NotebookWebApplication(web.Application):
 
         _template_path = settings_overrides.get(
             "template_path",
-            ipython_app.template_file_path,
+            jupyter_app.template_file_path,
         )
         if isinstance(_template_path, py3compat.string_types):
             _template_path = (_template_path,)
@@ -164,14 +196,28 @@ class NotebookWebApplication(web.Application):
             # reset the cache on server restart
             version_hash = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
 
+        if jupyter_app.ignore_minified_js:
+            log.warning("""The `ignore_minified_js` flag is deprecated and no 
+                longer works.  Alternatively use `npm run build:watch` when
+                working on the notebook's Javascript and LESS""")
+            warnings.warn("The `ignore_minified_js` flag is deprecated and will be removed in Notebook 6.0", DeprecationWarning)
+
+        now = utcnow()
+        
+        root_dir = contents_manager.root_dir
+        home = os.path.expanduser('~')
+        if root_dir.startswith(home + os.path.sep):
+            # collapse $HOME to ~
+            root_dir = '~' + root_dir[len(home):]
+
         settings = dict(
             # basics
             log_function=log_request,
             base_url=base_url,
             default_url=default_url,
             template_path=template_path,
-            static_path=ipython_app.static_file_path,
-            static_custom_path=ipython_app.static_custom_path,
+            static_path=jupyter_app.static_file_path,
+            static_custom_path=jupyter_app.static_custom_path,
             static_handler_class = FileFindHandler,
             static_url_prefix = url_path_join(base_url,'/static/'),
             static_handler_args = {
@@ -179,21 +225,26 @@ class NotebookWebApplication(web.Application):
                 'no_cache_paths': [url_path_join(base_url, 'static', 'custom')],
             },
             version_hash=version_hash,
-            ignore_minified_js=ipython_app.ignore_minified_js,
-
+            ignore_minified_js=jupyter_app.ignore_minified_js,
+            
             # rate limits
-            iopub_msg_rate_limit=ipython_app.iopub_msg_rate_limit,
-            iopub_data_rate_limit=ipython_app.iopub_data_rate_limit,
-            rate_limit_window=ipython_app.rate_limit_window,
+            iopub_msg_rate_limit=jupyter_app.iopub_msg_rate_limit,
+            iopub_data_rate_limit=jupyter_app.iopub_data_rate_limit,
+            rate_limit_window=jupyter_app.rate_limit_window,
 
+            # maximum request sizes - support saving larger notebooks
+            # tornado defaults are 100 MiB, we increase it to 0.5 GiB
+            max_body_size = 512 * 1024 * 1024,
+            max_buffer_size = 512 * 1024 * 1024,
+            
             # authentication
-            cookie_secret=ipython_app.cookie_secret,
+            cookie_secret=jupyter_app.cookie_secret,
             login_url=url_path_join(base_url,'/login'),
-            login_handler_class=ipython_app.login_handler_class,
-            logout_handler_class=ipython_app.logout_handler_class,
-            password=ipython_app.password,
+            login_handler_class=jupyter_app.login_handler_class,
+            logout_handler_class=jupyter_app.logout_handler_class,
+            password=jupyter_app.password,
             xsrf_cookies=True,
-            disable_check_xsrf=ipython_app.disable_check_xsrf,
+            disable_check_xsrf=jupyter_app.disable_check_xsrf,
 
             # managers
             kernel_manager=kernel_manager,
@@ -202,13 +253,16 @@ class NotebookWebApplication(web.Application):
             kernel_spec_manager=kernel_spec_manager,
             config_manager=config_manager,
 
-            # IPython stuff
-            jinja_template_vars=ipython_app.jinja_template_vars,
-            nbextensions_path=ipython_app.nbextensions_path,
-            websocket_url=ipython_app.websocket_url,
-            mathjax_url=ipython_app.mathjax_url,
-            config=ipython_app.config,
-            config_dir=ipython_app.config_dir,
+            # Jupyter stuff
+            started=now,
+            jinja_template_vars=jupyter_app.jinja_template_vars,
+            nbextensions_path=jupyter_app.nbextensions_path,
+            websocket_url=jupyter_app.websocket_url,
+            mathjax_url=jupyter_app.mathjax_url,
+            mathjax_config=jupyter_app.mathjax_config,
+            config=jupyter_app.config,
+            config_dir=jupyter_app.config_dir,
+            server_root_dir=root_dir,
             jinja2_env=env,
             terminals_available=False,  # Set later if terminals are available
         )
@@ -226,8 +280,10 @@ class NotebookWebApplication(web.Application):
         handlers.extend([(r"/login", settings['login_handler_class'])])
         handlers.extend([(r"/logout", settings['logout_handler_class'])])
         handlers.extend(load_handlers('files.handlers'))
+        handlers.extend(load_handlers('view.handlers'))
         handlers.extend(load_handlers('notebook.handlers'))
         handlers.extend(load_handlers('nbconvert.handlers'))
+        handlers.extend(load_handlers('bundler.handlers'))
         handlers.extend(load_handlers('kernelspecs.handlers'))
         handlers.extend(load_handlers('edit.handlers'))
         handlers.extend(load_handlers('services.api.handlers'))
@@ -239,23 +295,6 @@ class NotebookWebApplication(web.Application):
         handlers.extend(load_handlers('services.kernelspecs.handlers'))
         handlers.extend(load_handlers('services.security.handlers'))
         
-        # BEGIN HARDCODED WIDGETS HACK
-        # TODO: Remove on notebook 5.0
-        try:
-            import widgetsnbextension
-        except:
-            try:
-                import ipywidgets as widgets
-                handlers.append(
-                    (r"/nbextensions/widgets/(.*)", FileFindHandler, {
-                        'path': widgets.find_static_assets(),
-                        'no_cache_paths': ['/'], # don't cache anything in nbextensions
-                    }),
-                )
-            except:
-                app_log.warning('Widgets are unavailable. Please install widgetsnbextension or ipywidgets 4.0')
-        # END HARDCODED WIDGETS HACK
-        
         handlers.append(
             (r"/nbextensions/(.*)", FileFindHandler, {
                 'path': settings['nbextensions_path'],
@@ -289,6 +328,24 @@ class NotebookWebApplication(web.Application):
         return new_handlers
 
 
+class NotebookPasswordApp(JupyterApp):
+    """Set a password for the notebook server.
+
+    Setting a password secures the notebook server
+    and removes the need for token-based authentication.
+    """
+    
+    description = __doc__
+
+    def _config_file_default(self):
+        return os.path.join(self.config_dir, 'jupyter_notebook_config.json')
+
+    def start(self):
+        from .auth.security import set_password
+        set_password(config_file=self.config_file)
+        self.log.info("Wrote hashed password to %s" % self.config_file)
+
+
 class NbserverListApp(JupyterApp):
     version = __version__
     description="List currently running notebook servers."
@@ -339,6 +396,11 @@ flags['no-mathjax']=(
     """
 )
 
+flags['allow-root']=(
+    {'NotebookApp' : {'allow_root' : True}},
+    "Allow the notebook to be run from root user."
+)
+
 # Add notebook manager flags
 flags.update(boolean_flag('script', 'FileContentsManager.save_script',
                'DEPRECATED, IGNORED',
@@ -387,24 +449,28 @@ class NotebookApp(JupyterApp):
     
     subcommands = dict(
         list=(NbserverListApp, NbserverListApp.description.splitlines()[0]),
+        password=(NotebookPasswordApp, NotebookPasswordApp.description.splitlines()[0]),
     )
 
     _log_formatter_cls = LogFormatter
 
-    def _log_level_default(self):
+    @default('log_level')
+    def _default_log_level(self):
         return logging.INFO
 
-    def _log_datefmt_default(self):
+    @default('log_datefmt')
+    def _default_log_datefmt(self):
         """Exclude date from default date format"""
         return "%H:%M:%S"
     
-    def _log_format_default(self):
+    @default('log_format')
+    def _default_log_format(self):
         """override default log format to include time"""
         return u"%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s]%(end_color)s %(message)s"
 
     ignore_minified_js = Bool(False,
             config=True,
-            help='Use minified JS file or not, mainly use during dev to avoid JS recompilation', 
+            help='Deprecated: Use minified JS file or not, mainly use during dev to avoid JS recompilation', 
             )
 
     # file to be opened in the notebook server
@@ -438,6 +504,10 @@ class NotebookApp(JupyterApp):
         help="Set the Access-Control-Allow-Credentials: true header"
     )
     
+    allow_root = Bool(False, config=True, 
+        help="Whether to allow the user to run the notebook as root."
+    )
+
     default_url = Unicode('/tree', config=True,
         help="The default URL to redirect to from `/`"
     )
@@ -445,7 +515,9 @@ class NotebookApp(JupyterApp):
     ip = Unicode('localhost', config=True,
         help="The IP address the notebook server will listen on."
     )
-    def _ip_default(self):
+
+    @default('ip')
+    def _default_ip(self):
         """Return localhost if available, 127.0.0.1 otherwise.
         
         On some (horribly broken) systems, localhost cannot be bound.
@@ -454,18 +526,23 @@ class NotebookApp(JupyterApp):
         try:
             s.bind(('localhost', 0))
         except socket.error as e:
-            self.log.warn("Cannot bind to localhost, using 127.0.0.1 as default ip\n%s", e)
+            self.log.warning("Cannot bind to localhost, using 127.0.0.1 as default ip\n%s", e)
             return '127.0.0.1'
         else:
             s.close()
             return 'localhost'
 
-    def _ip_changed(self, name, old, new):
-        if new == u'*': self.ip = u''
+    @validate('ip')
+    def _valdate_ip(self, proposal):
+        value = proposal['value']
+        if value == u'*':
+            value = u''
+        return value
 
     port = Integer(8888, config=True,
         help="The port the notebook server will listen on."
     )
+
     port_retries = Integer(50, config=True,
         help="The number of additional ports to try if the specified port is not available."
     )
@@ -485,7 +562,9 @@ class NotebookApp(JupyterApp):
     cookie_secret_file = Unicode(config=True,
         help="""The file where the cookie secret is stored."""
     )
-    def _cookie_secret_file_default(self):
+
+    @default('cookie_secret_file')
+    def _default_cookie_secret_file(self):
         return os.path.join(self.runtime_dir, 'notebook_cookie_secret')
     
     cookie_secret = Bytes(b'', config=True,
@@ -497,12 +576,14 @@ class NotebookApp(JupyterApp):
         cookie_secret stored in plaintext (you can read the value from a file).
         """
     )
-    def _cookie_secret_default(self):
+    
+    @default('cookie_secret')
+    def _default_cookie_secret(self):
         if os.path.exists(self.cookie_secret_file):
             with io.open(self.cookie_secret_file, 'rb') as f:
                 return f.read()
         else:
-            secret = base64.encodestring(os.urandom(1024))
+            secret = encodebytes(os.urandom(1024))
             self._write_cookie_secret_file(secret)
             return secret
     
@@ -514,12 +595,12 @@ class NotebookApp(JupyterApp):
         try:
             os.chmod(self.cookie_secret_file, 0o600)
         except OSError:
-            self.log.warn(
+            self.log.warning(
                 "Could not set permissions on %s",
                 self.cookie_secret_file
             )
 
-    token = Unicode('<generated>', config=True,
+    token = Unicode('<generated>',
         help="""Token used for authenticating first-time connections to the server.
 
         When no password is enabled,
@@ -527,7 +608,7 @@ class NotebookApp(JupyterApp):
 
         Setting to an empty string disables authentication altogether, which is NOT RECOMMENDED.
         """
-    )
+    ).tag(config=True)
 
     one_time_token = Unicode(
         help="""One-time token used for opening a browser.
@@ -535,9 +616,10 @@ class NotebookApp(JupyterApp):
         Once used, this token cannot be used again.
         """
     )
-    
+
     _token_generated = True
 
+    @default('token')
     def _token_default(self):
         if self.password:
             # no token if password is enabled
@@ -547,7 +629,8 @@ class NotebookApp(JupyterApp):
             self._token_generated = True
             return binascii.hexlify(os.urandom(24)).decode('ascii')
 
-    def _token_changed(self, name, old, new):
+    @observe('token')
+    def _token_changed(self, change):
         self._token_generated = False
 
     password = Unicode(u'', config=True,
@@ -561,6 +644,17 @@ class NotebookApp(JupyterApp):
                       """
     )
 
+    password_required = Bool(False, config=True,
+                      help="""Forces users to use a password for the Notebook server.
+                      This is useful in a multi user environment, for instance when
+                      everybody in the LAN can access each other's machine though ssh.
+
+                      In such a case, server the notebook server on localhost is not secure
+                      since any user can connect to the notebook server via ssh.
+
+                      """
+    )
+
     disable_check_xsrf = Bool(False, config=True,
         help="""Disable cross-site-request-forgery protection
 
@@ -596,14 +690,19 @@ class NotebookApp(JupyterApp):
     webapp_settings = Dict(config=True,
         help="DEPRECATED, use tornado_settings"
     )
-    def _webapp_settings_changed(self, name, old, new):
-        self.log.warn("\n    webapp_settings is deprecated, use tornado_settings.\n")
-        self.tornado_settings = new
+
+    @observe('webapp_settings') 
+    def _update_webapp_settings(self, change):
+        self.log.warning("\n    webapp_settings is deprecated, use tornado_settings.\n")
+        self.tornado_settings = change['new']
     
     tornado_settings = Dict(config=True,
             help="Supply overrides for the tornado.web.Application that the "
                  "Jupyter notebook uses.")
     
+    terminado_settings = Dict(config=True,
+            help='Supply overrides for terminado. Currently only supports "shell_command".')
+
     cookie_options = Dict(config=True,
         help="Extra keyword arguments to pass to `set_secure_cookie`."
              " See tornado's set_secure_cookie docs for details."
@@ -630,9 +729,11 @@ class NotebookApp(JupyterApp):
         When disabled, equations etc. will appear as their untransformed TeX source.
         """
     )
-    def _enable_mathjax_changed(self, name, old, new):
+
+    @observe('enable_mathjax')
+    def _update_enable_mathjax(self, change):
         """set mathjax url to empty if mathjax is disabled"""
-        if not new:
+        if not change['new']:
             self.mathjax_url = u''
 
     base_url = Unicode('/', config=True,
@@ -641,16 +742,22 @@ class NotebookApp(JupyterApp):
                                Leading and trailing slashes can be omitted,
                                and will automatically be added.
                                ''')
-    def _base_url_changed(self, name, old, new):
-        if not new.startswith('/'):
-            self.base_url = '/'+new
-        elif not new.endswith('/'):
-            self.base_url = new+'/'
+
+    @validate('base_url')
+    def _update_base_url(self, proposal):
+        value = proposal['value']
+        if not value.startswith('/'):
+            value = '/' + value
+        elif not value.endswith('/'):
+            value = value + '/'
+        return value
     
     base_project_url = Unicode('/', config=True, help="""DEPRECATED use base_url""")
-    def _base_project_url_changed(self, name, old, new):
-        self.log.warn("base_project_url is deprecated, use base_url")
-        self.base_url = new
+
+    @observe('base_project_url')
+    def _update_base_project_url(self, change):
+        self.log.warning("base_project_url is deprecated, use base_url")
+        self.base_url = change['new']
 
     extra_static_paths = List(Unicode(), config=True,
         help="""Extra paths to search for serving static files.
@@ -667,7 +774,9 @@ class NotebookApp(JupyterApp):
     static_custom_path = List(Unicode(),
         help="""Path to search for custom.js, css"""
     )
-    def _static_custom_path_default(self):
+
+    @default('static_custom_path')
+    def _default_static_custom_path(self):
         return [
             os.path.join(d, 'custom') for d in (
                 self.config_dir,
@@ -709,36 +818,51 @@ class NotebookApp(JupyterApp):
         Should be in the form of an HTTP origin: ws[s]://hostname[:port]
         """
     )
+
     mathjax_url = Unicode("", config=True,
         help="""A custom url for MathJax.js.
         Should be in the form of a case-sensitive url to MathJax,
         for example:  /static/components/MathJax/MathJax.js
         """
     )
-    def _mathjax_url_default(self):
+
+    @default('mathjax_url')
+    def _default_mathjax_url(self):
         if not self.enable_mathjax:
             return u''
         static_url_prefix = self.tornado_settings.get("static_url_prefix", "static")
         return url_path_join(static_url_prefix, 'components', 'MathJax', 'MathJax.js')
     
-    def _mathjax_url_changed(self, name, old, new):
+    @observe('mathjax_url')
+    def _update_mathjax_url(self, change):
+        new = change['new']
         if new and not self.enable_mathjax:
             # enable_mathjax=False overrides mathjax_url
             self.mathjax_url = u''
         else:
             self.log.info("Using MathJax: %s", new)
 
+    mathjax_config = Unicode("TeX-AMS-MML_HTMLorMML-full,Safe", config=True,
+        help="""The MathJax.js configuration file that is to be used."""
+    )
+
+    @observe('mathjax_config')
+    def _update_mathjax_config(self, change):
+        self.log.info("Using MathJax configuration file: %s", change['new'])
+
     contents_manager_class = Type(
-        default_value=FileContentsManager,
+        default_value=LargeFileManager,
         klass=ContentsManager,
         config=True,
         help='The notebook manager class to use.'
     )
+
     kernel_manager_class = Type(
         default_value=MappingKernelManager,
         config=True,
         help='The kernel manager class to use.'
     )
+
     session_manager_class = Type(
         default_value=SessionManager,
         config=True,
@@ -786,7 +910,8 @@ class NotebookApp(JupyterApp):
 
     info_file = Unicode()
 
-    def _info_file_default(self):
+    @default('info_file')
+    def _default_info_file(self):
         info_file = "nbserver-%s.json" % os.getpid()
         return os.path.join(self.runtime_dir, info_file)
     
@@ -795,10 +920,12 @@ class NotebookApp(JupyterApp):
         DISABLED: use %pylab or %matplotlib in the notebook to enable matplotlib.
         """
     )
-    def _pylab_changed(self, name, old, new):
+
+    @observe('pylab')
+    def _update_pylab(self, change):
         """when --pylab is specified, display a warning and exit"""
-        if new != 'warn':
-            backend = ' %s' % new
+        if change['new'] != 'warn':
+            backend = ' %s' % change['new']
         else:
             backend = ''
         self.log.error("Support for specifying --pylab on the command line has been removed.")
@@ -811,21 +938,22 @@ class NotebookApp(JupyterApp):
         help="The directory to use for notebooks and kernels."
     )
 
-    def _notebook_dir_default(self):
+    @default('notebook_dir')
+    def _default_notebook_dir(self):
         if self.file_to_run:
             return os.path.dirname(os.path.abspath(self.file_to_run))
         else:
             return py3compat.getcwd()
 
-    def _notebook_dir_validate(self, value, trait):
+    @validate('notebook_dir')
+    def _notebook_dir_validate(self, proposal):
+        value = proposal['value']
         # Strip any trailing slashes
         # *except* if it's root
         _, path = os.path.splitdrive(value)
         if path == os.sep:
             return value
-
         value = value.rstrip(os.sep)
-
         if not os.path.isabs(value):
             # If we receive a non-absolute path, make it absolute.
             value = os.path.abspath(value)
@@ -833,9 +961,11 @@ class NotebookApp(JupyterApp):
             raise TraitError("No such notebook dir: %r" % value)
         return value
 
-    def _notebook_dir_changed(self, name, old, new):
+    @observe('notebook_dir')
+    def _update_notebook_dir(self, change):
         """Do a bit of validation of the notebook dir."""
         # setting App.notebook_dir implies setting notebook and kernel dirs as well
+        new = change['new']
         self.config.FileContentsManager.root_dir = new
         self.config.MappingKernelManager.root_dir = new
 
@@ -843,10 +973,12 @@ class NotebookApp(JupyterApp):
     server_extensions = List(Unicode(), config=True,
         help=("DEPRECATED use the nbserver_extensions dict instead")
     )
-    def _server_extensions_changed(self, name, old, new):
-        self.log.warning("server_extensions is deprecated, use nbserver_extensions")
-        self.server_extensions = new
     
+    @observe('server_extensions')
+    def _update_server_extensions(self, change):
+        self.log.warning("server_extensions is deprecated, use nbserver_extensions")
+        self.server_extensions = change['new']
+        
     nbserver_extensions = Dict({}, config=True,
         help=("Dict of Python modules to load as notebook server extensions."
               "Entry values can be used to enable and disable the loading of"
@@ -860,15 +992,15 @@ class NotebookApp(JupyterApp):
         help="Reraise exceptions encountered loading server extensions?",
     )
 
-    iopub_msg_rate_limit = Float(0, config=True, help="""(msg/sec)
+    iopub_msg_rate_limit = Float(1000, config=True, help="""(msgs/sec)
         Maximum rate at which messages can be sent on iopub before they are
         limited.""")
 
-    iopub_data_rate_limit = Float(0, config=True, help="""(bytes/sec)
+    iopub_data_rate_limit = Float(1000000, config=True, help="""(bytes/sec)
         Maximum rate at which messages can be sent on iopub before they are
         limited.""")
 
-    rate_limit_window = Float(1.0, config=True, help="""(sec) Time window used to 
+    rate_limit_window = Float(3, config=True, help="""(sec) Time window used to 
         check the message and data rate limits.""")
 
     def parse_command_line(self, argv=None):
@@ -914,7 +1046,6 @@ class NotebookApp(JupyterApp):
         self.config_manager = self.config_manager_class(
             parent=self,
             log=self.log,
-            config_dir=os.path.join(self.config_dir, 'nbconfig'),
         )
 
     def init_logging(self):
@@ -947,7 +1078,13 @@ class NotebookApp(JupyterApp):
         # ensure default_url starts with base_url
         if not self.default_url.startswith(self.base_url):
             self.default_url = url_path_join(self.base_url, self.default_url)
-        
+
+        if self.password_required and (not self.password):
+            self.log.critical("Notebook servers are configured to only be run with a password.")
+            self.log.critical("Hint: run the following command to set a password")
+            self.log.critical("\t$ python -m notebook.auth password")
+            sys.exit(1)
+
         self.web_app = NotebookWebApplication(
             self, self.kernel_manager, self.contents_manager,
             self.session_manager, self.kernel_spec_manager,
@@ -986,7 +1123,7 @@ class NotebookApp(JupyterApp):
                     self.log.info('The port %i is already in use, trying another port.' % port)
                     continue
                 elif e.errno in (errno.EACCES, getattr(errno, 'WSAEACCES', errno.EACCES)):
-                    self.log.warn("Permission to listen on port %i denied" % port)
+                    self.log.warning("Permission to listen on port %i denied" % port)
                     continue
                 else:
                     raise
@@ -1008,8 +1145,6 @@ class NotebookApp(JupyterApp):
             token = self.token if self._token_generated else '...'
             url = url_concat(url, {'token': token})
         return url
-        query = '?token=%s' % self.token if self.token else ''
-        return self._url(ip) + query
 
     @property
     def connection_url(self):
@@ -1023,10 +1158,10 @@ class NotebookApp(JupyterApp):
     def init_terminals(self):
         try:
             from .terminal import initialize
-            initialize(self.web_app, self.notebook_dir, self.connection_url)
+            initialize(self.web_app, self.notebook_dir, self.connection_url, self.terminado_settings)
             self.web_app.settings['terminals_available'] = True
         except ImportError as e:
-            log = self.log.debug if sys.platform == 'win32' else self.log.warn
+            log = self.log.debug if sys.platform == 'win32' else self.log.warning
             log("Terminals not available (error was %s)", e)
 
     def init_signal(self):
@@ -1110,9 +1245,28 @@ class NotebookApp(JupyterApp):
             # in the new traitlet
             if not modulename in self.nbserver_extensions:
                 self.nbserver_extensions[modulename] = True
-        
-        for modulename in sorted(self.nbserver_extensions):
-            if self.nbserver_extensions[modulename]:
+
+        # Load server extensions with ConfigManager.
+        # This enables merging on keys, which we want for extension enabling.
+        # Regular config loading only merges at the class level,
+        # so each level (user > env > system) clobbers the previous.
+        config_path = jupyter_config_path()
+        if self.config_dir not in config_path:
+            # add self.config_dir to the front, if set manually
+            config_path.insert(0, self.config_dir)
+        manager = ConfigManager(read_config_path=config_path)
+        section = manager.get(self.config_file_name)
+        extensions = section.get('NotebookApp', {}).get('nbserver_extensions', {})
+
+        for modulename, enabled in self.nbserver_extensions.items():
+            if modulename not in extensions:
+                # not present in `extensions` means it comes from Python config,
+                # so we need to add it.
+                # Otherwise, trust ConfigManager to have loaded it.
+                extensions[modulename] = enabled
+
+        for modulename, enabled in sorted(extensions.items()):
+            if enabled:
                 try:
                     mod = importlib.import_module(modulename)
                     func = getattr(mod, 'load_jupyter_server_extension', None)
@@ -1194,12 +1348,28 @@ class NotebookApp(JupyterApp):
         
         This method takes no arguments so all configuration and initialization
         must be done prior to calling this method."""
+
+        if not self.allow_root:
+            # check if we are running as root, and abort if it's not allowed
+            try:
+                uid = os.geteuid()
+            except AttributeError:
+                uid = -1 # anything nonzero here, since we can't check UID assume non-root
+            if uid == 0:
+                self.log.critical("Running as root is not recommended. Use --allow-root to bypass.")
+                self.exit(1)
+
         super(NotebookApp, self).start()
 
         info = self.log.info
         for line in self.notebook_info().split("\n"):
             info(line)
         info("Use Control-C to stop this server and shut down all kernels (twice to skip confirmation).")
+        if 'dev' in notebook.__version__:
+            info("Welcome to Project Jupyter! Explore the various tools available"
+                 " and their corresponding documentation. If you are interested"
+                 " in contributing to the platform, please visit the community"
+                 "resources section at http://jupyter.org/community.html.")
 
         self.write_server_info_file()
 
@@ -1207,7 +1377,7 @@ class NotebookApp(JupyterApp):
             try:
                 browser = webbrowser.get(self.browser or None)
             except webbrowser.Error as e:
-                self.log.warn('No web browser found: %s.' % e)
+                self.log.warning('No web browser found: %s.' % e)
                 browser = None
             
             if self.file_to_run:
@@ -1226,7 +1396,7 @@ class NotebookApp(JupyterApp):
                 b = lambda : browser.open(url_path_join(self.connection_url, uri),
                                           new=2)
                 threading.Thread(target=b).start()
-                
+
         if self.token and self._token_generated:
             # log full URL with generated token, so there's a copy/pasteable link
             # with auth info.
@@ -1235,7 +1405,7 @@ class NotebookApp(JupyterApp):
                 'Copy/paste this URL into your browser when you connect for the first time,',
                 'to login with a token:',
                 '    %s' % url_concat(self.connection_url, {'token': self.token}),
-             ]))
+            ]))
 
         self.io_loop = ioloop.IOLoop.current()
         if sys.platform.startswith('win'):
