@@ -1,8 +1,8 @@
 # coding: utf-8
 """Test the contents webservice API."""
 
-import base64
 from contextlib import contextmanager
+from functools import partial
 import io
 import json
 import os
@@ -18,7 +18,7 @@ from ..filecheckpoints import GenericFileCheckpoints
 from traitlets.config import Config
 from notebook.utils import url_path_join, url_escape, to_os_path
 from notebook.tests.launchnotebook import NotebookTestBase, assert_http_error
-from nbformat import read, write, from_dict
+from nbformat import write, from_dict
 from nbformat.v4 import (
     new_notebook, new_markdown_cell,
 )
@@ -26,6 +26,12 @@ from nbformat import v2
 from ipython_genutils import py3compat
 from ipython_genutils.tempdir import TemporaryDirectory
 
+try: #PY3
+    from base64 import encodebytes, decodebytes
+except ImportError: #PY2
+    from base64 import encodestring as encodebytes, decodestring as decodebytes
+
+
 def uniq_stable(elems):
     """uniq_stable(elems) -> list
 
@@ -146,7 +152,7 @@ class APITest(NotebookTestBase):
         return u'%s text file' % name
     
     def to_os_path(self, api_path):
-        return to_os_path(api_path, root=self.notebook_dir.name)
+        return to_os_path(api_path, root=self.notebook_dir)
     
     def make_dir(self, api_path):
         """Create a directory at api_path"""
@@ -190,32 +196,32 @@ class APITest(NotebookTestBase):
     
     def isdir(self, api_path):
         return os.path.isdir(self.to_os_path(api_path))
-    
-    def setUp(self):
 
+    def setUp(self):
         for d in (self.dirs + self.hidden_dirs):
             self.make_dir(d)
+            self.addCleanup(partial(self.delete_dir, d))
 
         for d, name in self.dirs_nbs:
             # create a notebook
             nb = new_notebook()
-            self.make_nb(u'{}/{}.ipynb'.format(d, name), nb)
-            
+            nbname = u'{}/{}.ipynb'.format(d, name)
+            self.make_nb(nbname, nb)
+            self.addCleanup(partial(self.delete_file, nbname))
+
             # create a text file
             txt = self._txt_for_name(name)
-            self.make_txt(u'{}/{}.txt'.format(d, name), txt)
-            
-            # create a binary file
+            txtname = u'{}/{}.txt'.format(d, name)
+            self.make_txt(txtname, txt)
+            self.addCleanup(partial(self.delete_file, txtname))
+
             blob = self._blob_for_name(name)
-            self.make_blob(u'{}/{}.blob'.format(d, name), blob)
+            blobname = u'{}/{}.blob'.format(d, name)
+            self.make_blob(blobname, blob)
+            self.addCleanup(partial(self.delete_file, blobname))
 
         self.api = API(self.request)
 
-    def tearDown(self):
-        for dname in (list(self.top_level_dirs) + self.hidden_dirs):
-            self.delete_dir(dname)
-        self.delete_file('inroot.ipynb')
-
     def test_list_notebooks(self):
         nbs = notebooks_only(self.api.list().json())
         self.assertEqual(len(nbs), 1)
@@ -286,6 +292,24 @@ class APITest(NotebookTestBase):
             self.assertIn('content', nb)
             self.assertEqual(nb['content'], None)
 
+    def test_get_nb_invalid(self):
+        nb = {
+            'nbformat': 4,
+            'metadata': {},
+            'cells': [{
+                'cell_type': 'wrong',
+                'metadata': {},
+            }],
+        }
+        path = u'å b/Validate tést.ipynb'
+        self.make_txt(path, py3compat.cast_unicode(json.dumps(nb)))
+        model = self.api.read(path).json()
+        self.assertEqual(model['path'], path)
+        self.assertEqual(model['type'], 'notebook')
+        self.assertIn('content', model)
+        self.assertIn('message', model)
+        self.assertIn("validation failed", model['message'].lower())
+
     def test_get_contents_no_such_file(self):
         # Name that doesn't exist - should be a 404
         with assert_http_error(404):
@@ -320,7 +344,7 @@ class APITest(NotebookTestBase):
             self.assertEqual(model['format'], 'base64')
             self.assertEqual(model['type'], 'file')
             self.assertEqual(
-                base64.decodestring(model['content'].encode('ascii')),
+                decodebytes(model['content'].encode('ascii')),
                 self._blob_for_name(name),
             )
 
@@ -415,7 +439,7 @@ class APITest(NotebookTestBase):
 
     def test_upload_b64(self):
         body = b'\xFFblob'
-        b64body = base64.encodestring(body).decode('ascii')
+        b64body = encodebytes(body).decode('ascii')
         model = {
             'content' : b64body,
             'format'  : 'base64',
@@ -430,7 +454,7 @@ class APITest(NotebookTestBase):
         self.assertEqual(model['type'], 'file')
         self.assertEqual(model['path'], path)
         self.assertEqual(model['format'], 'base64')
-        decoded = base64.decodestring(model['content'].encode('ascii'))
+        decoded = decodebytes(model['content'].encode('ascii'))
         self.assertEqual(decoded, body)
 
     def test_upload_v2(self):
