@@ -8,6 +8,8 @@ Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-16%3A-
 
 import json
 import logging
+from textwrap import dedent
+
 from tornado import gen, web
 from tornado.concurrent import Future
 from tornado.ioloop import IOLoop
@@ -29,7 +31,7 @@ class MainKernelHandler(APIHandler):
     def get(self):
         km = self.kernel_manager
         kernels = yield gen.maybe_future(km.list_kernels())
-        self.finish(json.dumps(kernels))
+        self.finish(json.dumps(kernels, default=date_default))
 
     @json_errors
     @web.authenticated
@@ -49,7 +51,7 @@ class MainKernelHandler(APIHandler):
         location = url_path_join(self.base_url, 'api', 'kernels', url_escape(kernel_id))
         self.set_header('Location', location)
         self.set_status(201)
-        self.finish(json.dumps(model))
+        self.finish(json.dumps(model, default=date_default))
 
 
 class KernelHandler(APIHandler):
@@ -60,7 +62,7 @@ class KernelHandler(APIHandler):
         km = self.kernel_manager
         km._check_kernel_id(kernel_id)
         model = km.kernel_model(kernel_id)
-        self.finish(json.dumps(model))
+        self.finish(json.dumps(model, default=date_default))
 
     @json_errors
     @web.authenticated
@@ -91,7 +93,7 @@ class KernelActionHandler(APIHandler):
                 self.set_status(500)
             else:
                 model = km.kernel_model(kernel_id)
-                self.write(json.dumps(model))
+                self.write(json.dumps(model, default=date_default))
         self.finish()
 
 
@@ -108,11 +110,11 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
     @property
     def iopub_msg_rate_limit(self):
-        return self.settings.get('iopub_msg_rate_limit', None)
+        return self.settings.get('iopub_msg_rate_limit', 0)
 
     @property
     def iopub_data_rate_limit(self):
-        return self.settings.get('iopub_data_rate_limit', None)
+        return self.settings.get('iopub_data_rate_limit', 0)
 
     @property
     def rate_limit_window(self):
@@ -229,7 +231,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             """Don't wait forever for the kernel to reply"""
             if future.done():
                 return
-            self.log.warn("Timeout waiting for kernel_info reply from %s", self.kernel_id)
+            self.log.warning("Timeout waiting for kernel_info reply from %s", self.kernel_id)
             future.set_result({})
         loop = IOLoop.current()
         loop.add_timeout(loop.time() + self.kernel_info_timeout, give_up)
@@ -258,6 +260,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
     
     def open(self, kernel_id):
         super(ZMQChannelsHandler, self).open()
+        self.kernel_manager.notify_connect(kernel_id)
         try:
             self.create_stream()
         except web.HTTPError as e:
@@ -283,10 +286,10 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             msg = json.loads(msg)
         channel = msg.pop('channel', None)
         if channel is None:
-            self.log.warn("No channel specified, assuming shell: %s", msg)
+            self.log.warning("No channel specified, assuming shell: %s", msg)
             channel = 'shell'
         if channel not in self.channels:
-            self.log.warn("No such channel: %r", channel)
+            self.log.warning("No such channel: %r", channel)
             return
         stream = self.channels[channel]
         self.session.send(stream, msg)
@@ -296,9 +299,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         msg = self.session.deserialize(fed_msg_list)
         parent = msg['parent_header']
         def write_stderr(error_message):
-            self.log.warn(error_message)
+            self.log.warning(error_message)
             msg = self.session.msg("stream",
-                content={"text": error_message, "name": "stderr"},
+                content={"text": error_message + '\n', "name": "stderr"},
                 parent=parent
             )
             msg['channel'] = 'iopub'
@@ -306,6 +309,16 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             
         channel = getattr(stream, 'channel', None)
         msg_type = msg['header']['msg_type']
+
+        if channel == 'iopub' and msg_type == 'status' and msg['content'].get('execution_state') == 'idle':
+            # reset rate limit counter on status=idle,
+            # to avoid 'Run All' hitting limits prematurely.
+            self._iopub_window_byte_queue = []
+            self._iopub_window_msg_count = 0
+            self._iopub_window_byte_count = 0
+            self._iopub_msgs_exceeded = False
+            self._iopub_data_exceeded = False
+
         if channel == 'iopub' and msg_type not in {'status', 'comm_open', 'execute_input'}:
             
             # Remove the counts queued for removal.
@@ -320,7 +333,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                     # This part of the queue hasn't be reached yet, so we can
                     # abort the loop.
                     break
-                    
+
             # Increment the bytes and message count
             self._iopub_window_msg_count += 1
             byte_count = sum([len(x) for x in msg_list])
@@ -336,39 +349,45 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             data_rate = float(self._iopub_window_byte_count) / self.rate_limit_window
             
             # Check the msg rate
-            if self.iopub_msg_rate_limit is not None and msg_rate > self.iopub_msg_rate_limit and self.iopub_msg_rate_limit > 0:
+            if self.iopub_msg_rate_limit > 0 and msg_rate > self.iopub_msg_rate_limit:
                 if not self._iopub_msgs_exceeded:
                     self._iopub_msgs_exceeded = True
-                    write_stderr("""iopub message rate exceeded.  The
-                    notebook server will temporarily stop sending iopub
-                    messages to the client in order to avoid crashing it.
+                    write_stderr(dedent("""\
+                    IOPub message rate exceeded.
+                    The notebook server will temporarily stop sending output
+                    to the client in order to avoid crashing it.
                     To change this limit, set the config variable
-                    `--NotebookApp.iopub_msg_rate_limit`.""")
-                    return
+                    `--NotebookApp.iopub_msg_rate_limit`."""))
             else:
-                if self._iopub_msgs_exceeded:
+                # resume once we've got some headroom below the limit
+                if self._iopub_msgs_exceeded and msg_rate < (0.8 * self.iopub_msg_rate_limit):
                     self._iopub_msgs_exceeded = False
                     if not self._iopub_data_exceeded:
-                        self.log.warn("iopub messages resumed")
-    
+                        self.log.warning("iopub messages resumed")
+
             # Check the data rate
-            if self.iopub_data_rate_limit is not None and data_rate > self.iopub_data_rate_limit and self.iopub_data_rate_limit > 0:
+            if self.iopub_data_rate_limit > 0 and data_rate > self.iopub_data_rate_limit:
                 if not self._iopub_data_exceeded:
                     self._iopub_data_exceeded = True
-                    write_stderr("""iopub data rate exceeded.  The
-                    notebook server will temporarily stop sending iopub
-                    messages to the client in order to avoid crashing it.
+                    write_stderr(dedent("""\
+                    IOPub data rate exceeded.
+                    The notebook server will temporarily stop sending output
+                    to the client in order to avoid crashing it.
                     To change this limit, set the config variable
-                    `--NotebookApp.iopub_data_rate_limit`.""")
-                    return
+                    `--NotebookApp.iopub_data_rate_limit`."""))
             else:
-                if self._iopub_data_exceeded:
+                # resume once we've got some headroom below the limit
+                if self._iopub_data_exceeded and data_rate < (0.8 * self.iopub_data_rate_limit):
                     self._iopub_data_exceeded = False
                     if not self._iopub_msgs_exceeded:
-                        self.log.warn("iopub messages resumed")
+                        self.log.warning("iopub messages resumed")
         
             # If either of the limit flags are set, do not send the message.
             if self._iopub_msgs_exceeded or self._iopub_data_exceeded:
+                # we didn't send it, remove the current message from the calculus
+                self._iopub_window_msg_count -= 1
+                self._iopub_window_byte_count -= byte_count
+                self._iopub_window_byte_queue.pop(-1)
                 return
         super(ZMQChannelsHandler, self)._on_zmq_reply(stream, msg)
 
@@ -383,6 +402,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self._open_sessions.pop(self.session_key)
         km = self.kernel_manager
         if self.kernel_id in km:
+            km.notify_disconnect(self.kernel_id)
             km.remove_restart_callback(
                 self.kernel_id, self.on_kernel_restarted,
             )
